import 'package:socket_io_client/socket_io_client.dart' as IO;
import 'package:logging/logging.dart';

export 'package:socket_io_client/socket_io_client.dart';

typedef dynamic EventHandler<T>(T data);

class ApexSocket {
  final log = Logger('ApexSocket');

  static bool _isInitialized = false;
  static bool _isConnected = false;
  static IO.Socket socket;

  int _reconnectDelay = 2;

  bool get isInitialized => _isInitialized;

  bool get isConnected => socket != null && _isConnected;

  static final ApexSocket shared = ApexSocket._();

  ApexSocket._();

  static int port;
  static String url;
  static IO.OptionBuilder options;

  ApexSocket.init(String url, int port, IO.OptionBuilder options)
      : assert(url != null && Uri.parse(url).isAbsolute,
            '$url must be a valid url.'),
        assert(port >= 0 && port <= 65535,
            '$port must be a number between 0 and 65535.'),
        assert(!_isInitialized, 'ApexSocket has been initialized once.'),
        assert(options != null, 'Options must not be null') {
    ApexSocket.url = url;
    ApexSocket.port = port;
    ApexSocket.options = options;
    _isInitialized = true;
  }

  void createSocket({
    bool recreate = false,
    EventHandler onConnect,
    EventHandler onConnectError,
    EventHandler onError,
    EventHandler onDisconnect,
    EventHandler onReconnect,
  }) {
    assert(_isInitialized, 'ApexSocket is not initialized');
    assert(options != null, 'Options must not be null');
    if (socket == null || recreate) {
      if (recreate) if (ApexSocket.socket != null) ApexSocket.socket.destroy();
      IO.Socket socket = IO.io(
          '${ApexSocket.url}:${ApexSocket.port}',
          options
              .setReconnectionDelay(_reconnectDelay)
              .setTransports(['websocket']).build());

      socket.onConnect((data) {
        _isConnected = true;
        if (onConnect != null) return onConnect(data);
      });

      socket.onConnectError(onConnectError);

      socket.onError(onError);

      socket.onDisconnect((data) {
        _isConnected = false;
        if (onDisconnect != null) return onDisconnect(data);
      });

      socket.onReconnect((data) {
        if (_reconnectDelay < 64) {
          _reconnectDelay *= 2;
        } else {
          _reconnectDelay = 64;
        }
        // TODO : needs testing
        socket.io.options = options.setReconnectionDelay(_reconnectDelay).build();
        if (onReconnect != null) return onReconnect(data);
      });

      ApexSocket.socket = socket;
    }

    if (socket.connected)
      socket.connect();
    else
      log.severe('already connected');
  }
}
